﻿#region License
// Copyright 2010 Buu Nguyen (http://www.buunguyen.net/blog)
// 
// Licensed under the Apache License, Version 2.0 (the "License"); 
// you may not use this file except in compliance with the License. 
// You may obtain a copy of the License at 
// 
// http://www.apache.org/licenses/LICENSE-2.0 
// 
// Unless required by applicable law or agreed to in writing, software 
// distributed under the License is distributed on an "AS IS" BASIS, 
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
// See the License for the specific language governing permissions and 
// limitations under the License.
// 
// The latest version of this file can be found at http://combres.codeplex.com
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Xml.Linq;

namespace Combres.Binders
{
    /// <summary>
    /// A binder which converts <see cref="IEnumerable{T}"/> to a <see cref="IDictionary{TKey,TValue}"/>.
    /// Does not support nested parameters.
    /// </summary>
    /// <example>
    /// <code>
    /// var map = binder.Bind[IDictionary[string, string]](parameters);
    /// string value1 = map["key1"];
    /// string value2 = map["key2"];
    /// </code>
    /// </example>
    public sealed class ConvertToMapBinder : IObjectBinder
    {
        /// <inheritdoc cref="IObjectBinder.Bind{T}" />
        public T Bind<T>(IEnumerable<XElement> parameters)
        {
            var pairs = GetKeyValuePairs(parameters);
            return (T)(object)pairs.ToDictionary(p => p.Key, p => p.Value);
        }

        /// <inheritdoc cref="IObjectBinder.Bind" />
        public void Bind(IEnumerable<XElement> parameters, object instance)
        {
            var map = instance as IDictionary<string, string>;
            if (map == null)
            {
                throw new ArgumentException("instance must be of type IDictionary<string, string>");
            }
            var pairs = GetKeyValuePairs(parameters);
            foreach (var pair in pairs)
            {
                map.Add(pair);
            }
        }

        private static IEnumerable<KeyValuePair<string, string>> GetKeyValuePairs(
            IEnumerable<XElement> parameters)
        {
            return from param in parameters
                   select new KeyValuePair<string, string>(
                       param.Attr<string>("name"), param.Attr<string>("value"));
        }
    }
}
